#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

def grid(d1min, d1max, d2min, d2max, n_points1, n_points2, replicates=1, logscale=True, include_zero=False):
    replicates = int(replicates)

    if logscale:
        d1 = np.logspace(np.log10(d1min), np.log10(d1max), num=n_points1)
        d2 = np.logspace(np.log10(d2min), np.log10(d2max), num=n_points2)
    else:
        d1 = np.linspace(d1min, d1max, num=n_points1)
        d2 = np.linspace(d2min, d2max, num=n_points2)

    if include_zero and logscale:
        if d1min > 0:
            d1 = np.append(0, d1)
        if d2min > 0:
            d2 = np.append(0, d2)
    
    D1, D2 = np.meshgrid(d1,d2)
    D1 = D1.flatten()
    D2 = D2.flatten()

    D1 = np.hstack([D1,]*replicates)
    D2 = np.hstack([D2,]*replicates)

    return D1, D2

def grid_multi(dmin, dmax, npoints, logscale=True, include_zero=False):
    if not (len(dmin)==len(dmax) and len(dmin)==len(npoints)):
        return None
    doses = []
    for Dmin, Dmax, n in zip(dmin, dmax, npoints):
        if logscale:
            logDmin = np.log10(Dmin)
            logDmax = np.log10(Dmax)
            d = np.logspace(logDmin, logDmax, num=n)
        else:
            d = np.linspace(Dmin, Dmax, num=n)
        if include_zero and Dmin > 0:
            d = np.append(0, d)
        doses.append(d)
    dosegrid = np.meshgrid(*doses)
    
    if include_zero:
        total_length = np.prod([i+1 for i in npoints])
    else:
        total_length = np.prod(npoints)
    n = len(dmin)
    return_d = np.zeros((total_length, n))
    
    for i in range(n):
        return_d[:,i] = dosegrid[i].flatten()

    return return_d


def get_num_replicates(d1, d2):
    """Given 1d dose arrays d1 and d2, determine how many replicates of each unique combination are present

    Parameters:
    -----------
    d1 : array_like, float
        Doses of drug 1

    d2 : array_like, float
        Doses of drug 2

    Returns:
    -----------
    replicates : numpy.array
        Counts of each unique dose combination
    """
    return np.unique(np.asarray([d1,d2]), axis=1, return_counts=True)[1]

@DeprecationWarning
def remove_replicates(d1, d2):
    """Given 1d dose arrays d1 and d2, remove replicates. This is needed sometimes for plotting, since some plot functions expect a single d1, d2 -> E for each dose.

    Parameters:
    -----------
    d1 : array_like, float
        Doses of drug 1

    d2 : array_like, float
        Doses of drug 2

    Returns:
    -----------
    d1 : array_like, float
        Doses of drug 1 without replicates

    d2 : array_like, float
        Doses of drug 2 without replicates
    """
    d = np.asarray(list(set(zip(d1, d2))))
    return d[:,0], d[:,1]


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import warnings

from .. import utils
from ..single import MarginalLinear
from .nonparametric_base import DoseDependentModel


class HSA(DoseDependentModel):
    """Highest single agent (HSA)

    HSA says that any improvement a combination gives over the strongest single agent counts as synergy.
    """

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, **kwargs):
        
        d1 = np.asarray(d1)
        d2 = np.asarray(d2)
        E = np.asarray(E)
        super().fit(d1,d2,E, drug1_model=drug1_model, drug2_model=drug2_model, **kwargs)

        drug1_model = self.drug1_model
        drug2_model = self.drug2_model

        #self.synergy = []
        d1_min = np.min(d1)
        d2_min = np.min(d2)

        if (d1_min > 0 or d2_min > 0):
            warnings.warn("WARNING: HSA expects single-drug information for both drugs. min(d1)=%0.2e, min(d2)=%0.2e"%(d1_min, d2_min))
        

        self.drug1_model = drug1_model
        self.drug2_model = drug2_model

        E1_alone = drug1_model.E(d1)
        E2_alone = drug2_model.E(d2)

        self.reference = np.minimum(E1_alone, E2_alone)
        #self.synergy = np.minimum(E1_alone-E, E2_alone-E)
        self.synergy = self.reference - E
        
        self.synergy[(d1==0) | (d2==0)] = 0
        return self.synergy

    def _get_single_drug_classes(self):
        return MarginalLinear, None


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import norm

from .. import utils
from ..utils import plots, dose_tools



class ParametricModel(ABC):
    """Base class for paramterized synergy models, including MuSyC, Zimmer, GPDI, and BRAID.
    """
    def __init__(self):
        """Bounds for drug response parameters (for instance, given percent viability data, one might expect E to be bounded within (0,1)) can be set, or parameters can be explicitly set.
        """

        self.bounds = None
        self.fit_function = None
        self.jacobian_function = None
        
        self.converged = False

        self.sum_of_squares_residuals = None
        self.r_squared = None
        self.aic = None
        self.bic = None
        self.bootstrap_parameters = None

    def _score(self, d1, d2, E):
        """Calculate goodness of fit and model quality scores, including sum-of-squares residuals, R^2, Akaike Information Criterion (AIC), and Bayesian Information Criterion (BIC).

        If model is not yet paramterized, does nothing

        Called automatically during model.fit(d1, d2, E)

        Parameters
        ----------
        d1 : array_like
            Doses of drug 1
        
        d2 : array_like
            Doses of drug 2
        
        E : array_like
            Dose-response at doses d1 and d2
        """
        if (self._is_parameterized()):

            n_parameters = len(self._get_parameters())

            self.sum_of_squares_residuals = utils.residual_ss(d1, d2, E, self.E)
            self.r_squared = utils.r_squared(E, self.sum_of_squares_residuals)
            self.aic = utils.AIC(self.sum_of_squares_residuals, n_parameters, len(E))
            self.bic = utils.BIC(self.sum_of_squares_residuals, n_parameters, len(E))

    @abstractmethod
    def _get_parameters(self):
        """Returns all of the model's fit parameters

        Returns
        ----------
        parameters : list or tuple
            Model's parameters
        """
        pass

    @abstractmethod
    def get_parameters(self, confidence_interval=95):
        """Returns a dict of the model's parameters.

        When relevant, it will also return meaningful derived parameters. For instance, MuSyC has several parameters for E, but defines a synergy parameter beta as a function of E parameters. Thus, beta will also be included.
        
        If the model was fit to data with bootstrap_iterations > 0, this will also return the specified confidence interval.
        """
        pass

    @abstractmethod
    def summary(self, confidence_interval=95, tol=0.01):
        """Summarizes the model's synergy conclusions.

        For each synergy parameters, determines whether it indicates synergy or antagonism. When the model has been fit with bootstrap_parameters>0, the best fit, lower bound, and upper bound must all agree on synergy or antagonism.
    
        Parameters
        ----------
        confidence_interval : float, optional (default=95)
            If the model was fit() with bootstrap_parameters>0, confidence_interval will be used to get the upper and lower bounds. 

        tol : float, optional (default=0.01)
            Tolerance to determine synergy or antagonism. The parameter must exceed the threshold by at least tol (some parameters, like MuSyC's alpha which is antagonistic from 0 to 1, and synergistic from 1 to inf, will be log-scaled prior to comparison with tol)

        Returns
        ----------
        summary : str
            Tab-separated string. If the model has been bootstrapped, columns are [parameter, value, (lower,upper), synergism/antagonism]. If the model has not been bootstrapped, columns are [parameter, value, synergism/antagonism].
        """
        pass

    @abstractmethod
    def _get_single_drug_classes(self):
        """
        Returns
        -------
        default_single_class : class
            The default class type to use for single-drug models

        expected_single_superclass : class
            The required type for single-drug models. If a single-drug model is passed that is not an instance of this superclass, it will be re-instantiated using default_model
        """
        pass

    def _internal_fit(self, d, E, use_jacobian, verbose=True, **kwargs):
        """Internal method to fit the model to data (d,E)
        """
        try:
            if use_jacobian and self.jacobian_function is not None:
                popt, pcov = curve_fit(self.fit_function, d, E, bounds=self.bounds, jac=self.jacobian_function, **kwargs)
            else: 
                popt, pcov = curve_fit(self.fit_function, d, E, bounds=self.bounds, **kwargs)
            if True in np.isnan(popt):
                return None
            return self._transform_params_from_fit(popt)
        except Exception as err:
            if verbose:
                print("Exception during combination drug response fit: %s"%err)
            return None

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, use_jacobian = True, p0=None, bootstrap_iterations=0, seed=None, **kwargs):
        """Fit the model to data.

        Parameters
        ----------
        d1 : array_like
            Doses of drug 1
        
        d2 : array_like
            Doses of drug 2

        E : array_like
            Dose-response at doses d1 and d2

        drug1_model : single-drug-model, default=None
            Only used when p0 is None. Pre-defined, or fit, model (e.g., Hill()) of drug 1 alone. Parameters from this model are used to provide an initial guess of E0, E1, h1, and C1 for the 2D-model fit. If None (and p0 is None), then d1 and E will be masked where d2==min(d2), and used to fit a model for drug 1.

        drug2_model : single-drug-model, default=None
            Same as drug1_model, for drug 2.
        
        use_jacobian : bool, default=True
            If True, will use the Jacobian to help guide fit (ONLY MuSyC, Hill, and Hill_2P have Jacobian implemented yet). When the number
            of data points is less than a few hundred, this makes the fitting
            slower. However, it also improves the reliability with which a fit
            can be found. If drug1_model or drug2_model are None, use_jacobian will also be applied for their fits.

        p0 : tuple, default=None
            Initial guess for the parameters. If p0 is None (default), drug1_model and drug2_model will be used to obtain an initial guess. If they are also None, they will be fit to the data. If they fail to fit, the initial guess will be E0=max(E), Emax=min(E), h=1, C=median(d), and all synergy parameters are additive (i.e., at the boundary between antagonistic and synergistic)

        seed : int, default=None
            If not None, used as numpy.random.seed(start_seed) at the beginning of bootstrap resampling
        
        kwargs
            kwargs to pass to scipy.optimize.curve_fit()
        """

        if seed is not None: np.random.seed(seed)
        d1 = np.asarray(d1, dtype=np.float64)
        d2 = np.asarray(d2, dtype=np.float64)

        E = np.asarray(E)

        xdata = np.vstack((d1,d2))
        
        if 'p0' in kwargs:
            p0 = list(kwargs.get('p0'))
        else:
            p0 = None
        
        p0 = self._get_initial_guess(d1, d2, E, drug1_model, drug2_model, p0=p0)

        kwargs['p0']=p0
        
        with np.errstate(divide='ignore', invalid='ignore'):
            popt = self._internal_fit(xdata, E, use_jacobian, **kwargs)

        if popt is None:
            self._set_parameters(self._transform_params_from_fit(p0))
            self.converged = False
        else:
            self.converged = True
            self._set_parameters(popt)
            n_parameters = len(popt)
            n_samples = len(d1)
            if (n_samples - n_parameters - 1 > 0):
                self._score(d1, d2, E)
                kwargs['p0'] = self._transform_params_to_fit(popt)
                self._bootstrap_resample(d1, d2, E, use_jacobian, bootstrap_iterations, **kwargs)
    
    @abstractmethod
    def E(self, d1, d2):
        """Returns drug effect E at dose d1,d2 for a pre-defined or fitted model.

        Parameters
        ----------
        d1 : array_like
            Doses of drug 1
        
        d2 : array_like
            Doses of drug 2
        
        Returns
        ----------
        effect : array_like
            Evaluate's the model at doses d1 and d2
        """
        pass

    def _is_parameterized(self):
        """Returns False if any parameters are None or nan.

        Returns
        ----------
        is_parameterzed : bool
            True if all of the parameters are set. False if any are None or nan.
        """
        return not (None in self._get_parameters() or True in np.isnan(np.asarray(self._get_parameters())))

    @abstractmethod
    def _set_parameters(self, popt):
        """Internal method to set model parameters
        """
        pass

    @abstractmethod
    def _transform_params_from_fit(self, params):
        """Internal method to transform parameterss as needed.

        For instance, models that fit logh and logC must transform those to h and C
        """
        pass

    @abstractmethod
    def _transform_params_to_fit(self, params):
        """Internal method to transform parameterss as needed.

        For instance, models that fit logh and logC must transform from h and C
        """
        pass

    @abstractmethod
    def _get_initial_guess(self, d1, d2, E, drug1_model, drug2_model, p0=None):
        """Internal method to format and/or guess p0
        """
        pass

    
    def _bootstrap_resample(self, d1, d2, E, use_jacobian, bootstrap_iterations, seed=None, **kwargs):
        """Internal function to identify confidence intervals for parameters
        """

        if not self._is_parameterized(): return
        if not self.converged: return

        n_data_points = len(E)
        n_parameters = len(self._get_parameters())
        
        sigma_residuals = np.sqrt(self.sum_of_squares_residuals / (n_data_points - n_parameters))

        E_model = self.E(d1, d2)
        bootstrap_parameters = []

        xdata = np.vstack((d1,d2))

        #if start_seed is not None: np.random.seed(start_seed)
        for iteration in range(bootstrap_iterations):
            #if start_seed is not None: np.random.seed(start_seed + iteration)
            residuals_step = norm.rvs(loc=0, scale=sigma_residuals, size=n_data_points)

            # Add random noise to model prediction
            E_iteration = E_model + residuals_step

            # Fit noisy data
            with np.errstate(divide='ignore', invalid='ignore', over='ignore'):
                popt1 = self._internal_fit(xdata, E_iteration, verbose=False, use_jacobian=use_jacobian, **kwargs)
            
            if popt1 is not None:
                bootstrap_parameters.append(popt1)
        if len(bootstrap_parameters) > 0:
            self.bootstrap_parameters = np.vstack(bootstrap_parameters)
        else:
            self.bootstrap_parameters = None

    def get_parameter_range(self, confidence_interval=95):
        """Returns the lower bound and upper bound estimate for each parameter.

        Parameters:
        -----------
        confidence_interval : int, float, default=95
            % confidence interval to return. Must be between 0 and 100.
        """
        if not self._is_parameterized():
            return None
        if not self.converged:
            return None
        if confidence_interval < 0 or confidence_interval > 100:
            return None
        if self.bootstrap_parameters is None:
            return None

        lb = (100-confidence_interval)/2.
        ub = 100-lb
        return np.percentile(self.bootstrap_parameters, [lb, ub], axis=0)

    def plot_heatmap(self, d1, d2, cmap="YlGnBu", **kwargs):
        """Plots the model's effect, E(d1, d2) as a heatmap

        Parameters
        ----------
        d1 : array_like
            Doses of drug 1
        
        d2 : array_like
            Doses of drug 2
        
        kwargs
            kwargs passed to synergy.utils.plots.plot_heatmap()
        """
        if not self._is_parameterized():
            #raise ModelNotParameterizedError()
            return
        
        E = self.E(d1, d2)
        plots.plot_heatmap(d1, d2, E, cmap=cmap, **kwargs)

    def plot_residual_heatmap(self, d1, d2, E, cmap="RdBu", center_on_zero=True, **kwargs):
        """Plots the residuals of the fit model as a heatmap

        Parameters
        ----------
        d1 : array_like
            Doses of drug 1
        
        d2 : array_like
            Doses of drug 2

        E : array_like
            Observed drug effects

        cmap : string, default="RdBu"
            Colormap for the plot
        
        kwargs
            kwargs passed to synergy.utils.plots.plot_heatmap()
        """
        if not self._is_parameterized():
            #raise ModelNotParameterizedError()
            return
        
        Emodel = self.E(d1, d2)
        plots.plot_heatmap(d1, d2, E-Emodel, cmap=cmap, center_on_zero=center_on_zero, **kwargs)

    @abstractmethod
    def _reference_E(self, d1, d2):
        pass

    def plot_reference_heatmap(self, d1, d2, cmap="YlGnBu", **kwargs):
        if not self._is_parameterized():
            #raise ModelNotParameterizedError()
            return

        Ereference = self._reference_E(d1, d2)
        plots.plot_heatmap(d1, d2, Ereference, cmap=cmap, **kwargs)

    def plot_reference_surface(self, d1, d2, cmap="YlGnBu", **kwargs):
        if not self._is_parameterized():
            return
        Ereference = self._reference_E(d1, d2)
        plots.plot_surface_plotly(d1, d2, Ereference, cmap=cmap, **kwargs)

    def plot_delta_heatmap(self, d1, d2, cmap="PRGn", center_on_zero=True, **kwargs):
        if not self._is_parameterized():
            #raise ModelNotParameterizedError()
            return
        Ereference = self._reference_E(d1, d2)
        Emodel = self.E(d1, d2)
        plots.plot_heatmap(d1, d2, Ereference-Emodel, cmap=cmap, center_on_zero=center_on_zero, **kwargs)

    def plot_delta_surface(self, d1, d2, cmap="PRGn", center_on_zero=True, **kwargs):
        if not self._is_parameterized():
            return
        Ereference = self._reference_E(d1, d2)
        Emodel = self.E(d1, d2)
        plots.plot_surface_plotly(d1, d2, Ereference-Emodel, cmap=cmap, center_on_zero=center_on_zero, **kwargs)
    

    def plot_surface_plotly(self, d1, d2, cmap="YlGnBu", **kwargs):
        """Plots the model's effect, E(d1, d2) as a surface using synergy.utils.plots.plot_surface_plotly()

        Parameters
        ----------
        d1 : array_like
            Doses of drug 1
        
        d2 : array_like
            Doses of drug 2
        
        cmap : string, default="viridis"
            Colorscale for the plot

        kwargs
            kwargs passed to synergy.utils.plots.plot_heatmap()
        """
        if not self._is_parameterized():
            #raise ModelNotParameterizedError()
            return
        
        # d1 and d2 may come from data, and have replicates. This would cause problems with surface plots (replicates in scatter_points are fine, but replicates in the surface itself are not)
        #d1, d2 = dose_tools.remove_replicates(d1, d2)
        E = self.E(d1, d2)
        plots.plot_surface_plotly(d1, d2, E, cmap=cmap, **kwargs)


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..single import Hill
from .nonparametric_base import DoseDependentHigher

class Loewe(DoseDependentHigher):
    """Loewe Additivity Synergy
    
    Loewe's model of drug combination additivity expects a linear tradeoff, such that withholding X parts of drug 1 can be compensated for by adding Y parts of drug 2. In Loewe's model, X and Y are constant (e.g., withholding 5X parts of drug 1 will be compensated for with 5Y parts of drug 2.)

    synergy : array_like, float
        [0,1)=synergism, (1,inf)=antagonism
    """
    
    def fit(self, d, E, single_models=None, **kwargs):
        d = np.asarray(d)
        E = np.asarray(E)
        n = d.shape[1]
        
        super().fit(d, E, single_models=single_models, **kwargs)
        single_models=self.single_models

        d_singles = d*0 # The dose of each drug that alone achieves E
        with np.errstate(divide='ignore', invalid='ignore'):
            for i in range(n):
                d_singles[:,i] = single_models[i].E_inv(E)

            self.synergy = (d/d_singles).sum(axis=1)

        # Ensure all single-drug Loewe scores are 1
        for i in range(n):
                # Mask where all other drugs are minimum (ideally 0)
                mask = d[:,i]>=0 # This should always be true
                for j in range(n):
                    if i==j: continue
                    mask = mask & (d[:,j]==np.min(d[:,j]))
                mask = np.where(mask)
                self.synergy[mask] = 1

        return self.synergy

    def _get_single_drug_classes(self):
        return Hill, None

    def plotly_isosurfaces(self, drug_axes=[0,1,2], other_drug_slices=None, cmap="PRGn", neglog=True, **kwargs):
        super().plotly_isosurfaces(drug_axes=drug_axes, other_drug_slices=other_drug_slices, cmap=cmap, neglog=neglog, **kwargs)


__version__="0.5.1"
__author__="David J. Wooten"
__citation__="Wooten, DJ and Albert, RA. synergy - A Python library for calculating drug combination synergy. (2020) Bioinformatics. https://doi.org/10.1093/bioinformatics/btaa826"

from . import combination
from . import single
from . import utils


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from scipy.optimize import curve_fit
from scipy.stats import norm
import numpy as np
from .. import utils

class ParameterizedModel1D:
    def __init__(self):
        self.bounds = None
        self.fit_function = None
        self.jacobian_function = None
        
        self.converged = False
        self._fit = False

        self.sum_of_squares_residuals = None
        self.r_squared = None
        self.aic = None
        self.bic = None
        self.bootstrap_parameters = None
    
    def _internal_fit(self, d, E, use_jacobian, verbose=True, **kwargs):
        """Internal method to fit the model to data (d,E)
        """
        try:
        #if True:
            if use_jacobian:
                popt, pcov = curve_fit(self.fit_function, d, E, bounds=self.bounds, jac=self.jacobian_function, **kwargs)
            else: 
                popt, pcov = curve_fit(self.fit_function, d, E, bounds=self.bounds, **kwargs)
            if True in np.isnan(popt):
                return None
            return self._transform_params_from_fit(popt)
        except Exception as err:
            if verbose:
                print("Exception during single drug response fit: %s"%err)
            return None

    def _get_initial_guess(self, d, E, p0=None):
        """Internal method to format and/or guess p0
        """
        return p0

    def _set_parameters(self, popt):
        """Internal method to set model parameters
        """
        pass

    def fit(self, d, E, use_jacobian=True, bootstrap_iterations=0, bootstrap_confidence_interval=95, **kwargs):
        """Fit the Hill equation to data. Fitting algorithm searches for h and C in a log-scale, but all bounds and guesses should be provided in a linear scale.

        Parameters
        ----------
        d : array_like
            Array of doses measured
        
        E : array_like
            Array of effects measured at doses d
        
        use_jacobian : bool, default=True
            If True, will use the Jacobian to help guide fit. When the number
            of data points is less than a few hundred, this makes the fitting
            slower. However, it also improves the reliability with which a fit
            can be found.
        
        kwargs
            kwargs to pass to scipy.optimize.curve_fit()
        """

        self._fit = True
        d = np.asarray(d)
        E = np.asarray(E)

        if 'p0' in kwargs:
            p0 = list(kwargs.get('p0'))
        else:
            p0 = None
        p0 = self._get_initial_guess(d, E, p0=p0)
        kwargs['p0']=p0
        
        with np.errstate(divide='ignore', invalid='ignore'):
            popt = self._internal_fit(d, E, use_jacobian, **kwargs)

        if popt is None:
            self.converged = False
            self._set_parameters(self._transform_params_from_fit(p0))
        else:
            self.converged = True
            self._set_parameters(popt)

            n_parameters = len(popt)
            n_samples = len(d)
            if (n_samples - n_parameters - 1 > 0):
                self._score(d, E)
                kwargs['p0'] = self._transform_params_to_fit(popt)
                self._bootstrap_resample(d, E, use_jacobian, bootstrap_iterations, bootstrap_confidence_interval, **kwargs)

    def E(self, d):
        """Evaluate this model at dose d.

        Parameters
        ----------
        d : array_like
            Doses to calculate effect at
        
        Returns
        ----------
        effect : array_like
            Evaluate's the model at dose in d
        """
        ret = 0*d
        ret[:] = np.nan
        return ret
        
    def E_inv(self, E):
        """Evaluate the inverse of this model.

        Parameters
        ----------
        E : array_like
            Effects to get the doses for
        
        Returns
        ----------
        doses : array_like
            Doses which achieve effects E using this model. Will return np.nan for effects outside of the model's effect range, or for non-invertable models
        """
        ret = 0*d
        ret[:] = np.nan
        return ret

    def _transform_params_from_fit(self, params):
        """Internal method to transform parameterss as needed.

        For instance, models that fit logh and logC must transform those to h and C
        """
        return params

    def _transform_params_to_fit(self, params):
        """Internal method to transform parameterss as needed.

        For instance, models that fit logh and logC must transform from h and C
        """
        return params

    def get_parameters(self):
        """Returns model parameters
        """
        return []

    def get_parameter_range(self, confidence_interval=95):
        """Returns the lower bound and upper bound estimate for each parameter.

        Parameters:
        -----------
        confidence_interval : int, float, default=95
            % confidence interval to return. Must be between 0 and 100.
        """
        if not self._is_parameterized():
            return None
        if not self.converged:
            return None
        if confidence_interval < 0 or confidence_interval > 100:
            return None
        if self.bootstrap_parameters is None:
            return None

        lb = (100-confidence_interval)/2.
        ub = 100-lb
        return np.percentile(self.bootstrap_parameters, [lb, ub], axis=0)

    def _is_parameterized(self):
        """Internalized method to check all model parameters are set
        """
        return not (None in self.get_parameters() or True in np.isnan(np.asarray(self.get_parameters())))

    def _score(self, d, E):
        """Calculate goodness of fit and model quality scores, including sum-of-squares residuals, R^2, Akaike Information Criterion (AIC), and Bayesian Information Criterion (BIC).

        If model is not yet paramterized, does nothing

        Called automatically during model.fit(d1, d2, E)

        Parameters
        ----------
        d : array_like
            Doses
        
        E : array_like
            Measured dose-response at doses d
        """
        if (self._is_parameterized()):

            n_parameters = len(self.get_parameters())

            self.sum_of_squares_residuals = utils.residual_ss_1d(d, E, self.E)
            self.r_squared = utils.r_squared(E, self.sum_of_squares_residuals)
            self.aic = utils.AIC(self.sum_of_squares_residuals, n_parameters, len(E))
            self.bic = utils.BIC(self.sum_of_squares_residuals, n_parameters, len(E))

    def _bootstrap_resample(self, d, E, use_jacobian, bootstrap_iterations, confidence_interval, **kwargs):
        """Internal function to identify confidence intervals for parameters
        """

        if not self._is_parameterized(): return
        if not self.converged: return

        n_data_points = len(E)
        n_parameters = len(self.get_parameters())
        
        sigma_residuals = np.sqrt(self.sum_of_squares_residuals / (n_data_points - n_parameters))

        E_model = self.E(d)
        bootstrap_parameters = []

        for iteration in range(bootstrap_iterations):
            residuals_step = norm.rvs(loc=0, scale=sigma_residuals, size=n_data_points)

            # Add random noise to model prediction
            E_iteration = E_model + residuals_step

            # Fit noisy data
            with np.errstate(divide='ignore', invalid='ignore'):
                popt1 = self._internal_fit(d, E_iteration, verbose=False, use_jacobian=use_jacobian, **kwargs)
            
            if popt1 is not None:
                bootstrap_parameters.append(popt1)
        
        if len(bootstrap_parameters) > 0:
            self.bootstrap_parameters = np.vstack(bootstrap_parameters)
        else:
            self.bootstrap_parameters = None

    def is_fit(self):
        return self._is_parameterized()
        #return self._fit


# Parametric Models
from .musyc import MuSyC

# Nonparametric models
from .bliss import Bliss
from .hsa import HSA
from .loewe import Loewe
from .combination_index import CombinationIndex
from .schindler import Schindler

######################

# No higher order synergy models described

# Parametric Models
#from .zimmer import Zimmer # They do pairwise calculations of 2-drug combos
#from .braid import BRAID

# Nonparametric models
#from .zero_interaction_potency import ZIP # This could almost certainly be done by averaging over 3+ Hill equation slices, but isn't explicitly described


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .. import utils

from ..single import Hill
from .parametric_base import ParametricModel

class BRAID(ParametricModel):
    """BRAID synergy (doi:10.1038/srep25523).
    
    kappa and delta are the BRAID synergy parameters, though E3 is related to how much more effective the combination is than either drug alone.

    Parameters
    ----------
    [X]_bounds : tuple
        Upper and lower bounds for each parameter to constrain fits.

    variant : str , default="kappa"
        Options "kappa", "delta", "both". BRAID has model versions that fit synergy using the parameter "kappa", the parameter "delta", or both. The standard version only fits kappa, but the other variants are available.

    """
    def __init__(self, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf),  \
            E0_bounds=(-np.inf,np.inf), E1_bounds=(-np.inf,np.inf), \
            E2_bounds=(-np.inf,np.inf), E3_bounds=(-np.inf,np.inf), \
            C1_bounds=(0,np.inf), C2_bounds=(0,np.inf), kappa_bounds=(-2, np.inf), delta_bounds=(0, np.inf), E0=None, E1=None, E2=None, E3=None, h1=None, h2=None, C1=None, C2=None, kappa=None, delta=None, variant="kappa"):

        super().__init__()
        self.E0_bounds = E0_bounds
        self.E1_bounds = E1_bounds
        self.E2_bounds = E2_bounds
        self.E3_bounds = E3_bounds
        self.C1_bounds = C1_bounds
        self.C2_bounds = C2_bounds
        self.h1_bounds = h1_bounds
        self.h2_bounds = h2_bounds
        self.kappa_bounds = kappa_bounds
        self.delta_bounds = delta_bounds

        with np.errstate(divide='ignore'):
            self.logh1_bounds = (np.log(h1_bounds[0]), np.log(h1_bounds[1]))
            self.logC1_bounds = (np.log(C1_bounds[0]), np.log(C1_bounds[1]))
            self.logh2_bounds = (np.log(h2_bounds[0]), np.log(h2_bounds[1]))
            self.logC2_bounds = (np.log(C2_bounds[0]), np.log(C2_bounds[1]))
            self.logdelta_bounds = (np.log(delta_bounds[0]), np.log(delta_bounds[1]))

        self.E0 = E0
        self.E1 = E1
        self.E2 = E2
        self.E3 = E3
        self.h1 = h1
        self.h2 = h2
        self.C1 = C1
        self.C2 = C2
        self.kappa = kappa
        self.delta = delta

        self.variant = variant
        if variant == "kappa":
            self.fit_function = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, kappa: self._model(d[0], d[1], E0, E1, E2, E3, np.exp(logh1), np.exp(logh2), np.exp(logC1), np.exp(logC2), kappa, 1)

            self.bounds = tuple(zip(self.E0_bounds, self.E1_bounds, self.E2_bounds, self.E3_bounds, self.logh1_bounds, self.logh2_bounds, self.logC1_bounds, self.logC2_bounds, self.kappa_bounds))
        
        elif variant == "delta":
            self.fit_function = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, logdelta: self._model(d[0], d[1], E0, E1, E2, E3, np.exp(logh1), np.exp(logh2), np.exp(logC1), np.exp(logC2), 0, np.exp(logdelta))

            self.bounds = tuple(zip(self.E0_bounds, self.E1_bounds, self.E2_bounds, self.E3_bounds, self.logh1_bounds, self.logh2_bounds, self.logC1_bounds, self.logC2_bounds, self.logdelta_bounds))
        
        elif variant == "both":
            self.fit_function = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, kappa, logdelta: self._model(d[0], d[1], E0, E1, E2, E3, np.exp(logh1), np.exp(logh2), np.exp(logC1), np.exp(logC2), kappa, np.exp(logdelta))

            self.bounds = tuple(zip(self.E0_bounds, self.E1_bounds, self.E2_bounds, self.E3_bounds, self.logh1_bounds, self.logh2_bounds, self.logC1_bounds, self.logC2_bounds, self.kappa_bounds, self.logdelta_bounds))

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, use_jacobian = True, p0=None, bootstrap_iterations=0, **kwargs):
        super().fit(d1, d2, E, drug1_model=drug1_model, drug2_model=drug2_model, use_jacobian=use_jacobian, p0=p0, bootstrap_iterations=bootstrap_iterations, **kwargs)

        # The way BRAID fits E3 uses a max() function that can actually make E3 not matter at all, over a potentially large range. Untreated, this can cause its uncertainty to explode. But E3 should be defined as the value that gives the greatest maximum E_range. Here we must manually fix it for bootstrapping.
        if self.bootstrap_parameters is not None and self.bootstrap_parameters.shape[0]>0:
            for i in range(self.bootstrap_parameters.shape[0]):
                E0, E1, E2, E3, = self.bootstrap_parameters[i,:4]
                delta_Es = [E1-E0, E2-E0, E3-E0]
                max_delta_E_index = np.argmax(np.abs(delta_Es))
                max_delta_E = delta_Es[max_delta_E_index]
                E3 = max_delta_E + E0
                self.bootstrap_parameters[i,3] = E3


    def _get_initial_guess(self, d1, d2, E, drug1_model, drug2_model, p0=None):
        
        # If there is no intial guess, use single-drug models to come up with intitial guess
        if p0 is None:
            # Sanitize single-drug models
            default_class, expected_superclass = self._get_single_drug_classes()

            drug1_model = utils.sanitize_single_drug_model(drug1_model, default_class, expected_superclass=expected_superclass, E0_bounds=self.E0_bounds, Emax_bounds=self.E1_bounds, h_bounds=self.h1_bounds, C_bounds=self.C1_bounds)

            drug2_model = utils.sanitize_single_drug_model(drug2_model, default_class, expected_superclass=expected_superclass, E0_bounds=self.E0_bounds, Emax_bounds=self.E2_bounds, h_bounds=self.h2_bounds, C_bounds=self.C2_bounds)

            # Fit the single drug models if they were not pre-fit by the user
            if not drug1_model.is_fit():
                mask = np.where(d2==min(d2))
                drug1_model.fit(d1[mask], E[mask])
            if not drug2_model.is_fit():
                mask = np.where(d1==min(d1))
                drug2_model.fit(d2[mask], E[mask])
            
            # Get initial guesses of E0, E1, E2, h1, h2, C1, and C2 from single-drug fits
            E0_1, E1, h1, C1 = drug1_model.get_parameters()
            E0_2, E2, h2, C2 = drug2_model.get_parameters()
            E0 = (E0_1+E0_2)/2
            
            E3 = E[(d1==max(d1)) & (d2==max(d2))]
            if len(E3)>0: E3 = np.mean(E3)
            else: E3 = np.min(E)

            if self.variant == "kappa":
                p0 = [E0, E1, E2, E3, h1, h2, C1, C2, 0]
            elif self.variant == "delta":
                p0 = [E0, E1, E2, E3, h1, h2, C1, C2, 1]
            elif self.variant == "both":
                p0 = [E0, E1, E2, E3, h1, h2, C1, C2, 0, 1]
            
        p0 = list(self._transform_params_to_fit(p0))
        utils.sanitize_initial_guess(p0, self.bounds)
        return p0

    def _transform_params_from_fit(self, params):
        if self.variant == "kappa":
            E0, E1, E2, E3, logh1, logh2, logC1, logC2, kappa = params
        elif self.variant == "delta":
            E0, E1, E2, E3, logh1, logh2, logC1, logC2, logdelta = params
            delta = np.exp(logdelta)
        elif self.variant == "both":
            E0, E1, E2, E3, logh1, logh2, logC1, logC2, kappa, logdelta = params
            delta = np.exp(logdelta)
        else:
            return None
        
        h1 = np.exp(logh1)
        h2 = np.exp(logh2)
        C1 = np.exp(logC1)
        C2 = np.exp(logC2)

        if self.variant == "kappa":
            return E0, E1, E2, E3, h1, h2, C1, C2, kappa
        elif self.variant == "delta":
            return E0, E1, E2, E3, h1, h2, C1, C2, delta
        elif self.variant == "both":
            return E0, E1, E2, E3, h1, h2, C1, C2, kappa, delta
        

    def _transform_params_to_fit(self, params):
        if self.variant == "kappa":
            E0, E1, E2, E3, h1, h2, C1, C2, kappa  = params
        elif self.variant == "delta":
            E0, E1, E2, E3, h1, h2, C1, C2, delta  = params
            logdelta = np.log(delta)
        elif self.variant == "both":
            E0, E1, E2, E3, h1, h2, C1, C2, kappa, delta  = params
            logdelta = np.log(delta)
        else: return None

        logh1 = np.log(h1)
        logh2 = np.log(h2)
        logC1 = np.log(C1)
        logC2 = np.log(C2)

        if self.variant == "kappa":
            return E0, E1, E2, E3, logh1, logh2, logC1, logC2, kappa
        elif self.variant == "delta":
            return E0, E1, E2, E3, logh1, logh2, logC1, logC2, logdelta
        elif self.variant == "both":
            return E0, E1, E2, E3, logh1, logh2, logC1, logC2, kappa, logdelta
        
    def _set_parameters(self, popt):
        if self.variant == "kappa":
            self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.kappa = popt
        elif self.variant == "delta":
            self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.delta = popt
        elif self.variant == "both":
            self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.kappa, self.delta = popt

        # E3 is, by definition, the value of E that gives the greatest delta_E. In fitting, the max() function can make E3 have no impact, leading to very sloppy output. Thus here we correct it by setting E3 to whichever E gives the greatest delta_E.

        delta_Es = [self.E1-self.E0, self.E2-self.E0, self.E3-self.E0]
        max_delta_E_index = np.argmax(np.abs(delta_Es))
        max_delta_E = delta_Es[max_delta_E_index]
        self.E3 = max_delta_E + self.E0
        

    def E(self, d1, d2):
        if not self._is_parameterized():
            # ERROR
            return 0

        if self.variant == "kappa":
            return self._model(d1, d2, self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.kappa, 1)
        elif self.variant == "delta":
            return self._model(d1, d2, self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, 0, self.delta)
        elif self.variant == "both":
            return self._model(d1, d2, self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.kappa, self.delta)
        
    def _reference_E(self, d1, d2):
        if not self._is_parameterized():
            return None
        return self._model(d1, d2, self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, 0, 1)
    
    def _model(self, d1, d2, E0, E1, E2, E3, h1, h2, C1, C2, kappa, delta):
        """
From the braidrm R package (https://rdrr.io/cran/braidrm/man/evalBRAIDrsm.html)

The parameters of this equation must satisfy h1>0, h2>0, delta>0, kappa>-2, sign(E3-E0)=sign(E1-E0)=sign(E2-E0), |E3-E0|>=|E1-E0|, and |E3-E0|>=|E2-E0|.
        """
        delta_Es = [E1-E0, E2-E0, E3-E0]
        max_delta_E_index = np.argmax(np.abs(delta_Es))
        max_delta_E = delta_Es[max_delta_E_index]

        h = np.sqrt(h1*h2)
        power = 1/(delta*h)
        
        D1 = (E1-E0)/max_delta_E * np.power(d1/C1,h1)/(1+(1-(E1-E0)/max_delta_E)*np.power(d1/C1,h1))
        
        D2 = (E2-E0)/max_delta_E*np.power(d2/C2,h2)/(1+(1-(E2-E0)/max_delta_E)*np.power(d2/C2,h2))
        
        D = np.power(D1,power) + np.power(D2,power) +kappa*np.sqrt(np.power(D1,power)*np.power(D2,power))
        
        return E0 + max_delta_E/(1+np.power(D,-delta*h))

    def _get_parameters(self):
        if self.variant == "kappa":
            return self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.kappa
        elif self.variant == "delta":
            return self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.delta
        elif self.variant == "both":
            return self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.kappa, self.delta

    def _get_single_drug_classes(self):
        return Hill, Hill

    def get_parameters(self, confidence_interval=95):
        if not self._is_parameterized():
            return None
        
        if self.converged and self.bootstrap_parameters is not None:
            parameter_ranges = self.get_parameter_range(confidence_interval=confidence_interval)
        else:
            parameter_ranges = None

        params = dict()
        params['E0'] = [self.E0, ]
        params['E1'] = [self.E1, ]
        params['E2'] = [self.E2, ]
        params['E3'] = [self.E3, ]
        params['h1'] = [self.h1, ]
        params['h2'] = [self.h2, ]
        params['C1'] = [self.C1, ]
        params['C2'] = [self.C2, ]
        if self.variant in ["kappa","both"]:
            params['kappa'] = [self.kappa, ]
        if self.variant in ["delta","both"]:
            params['delta'] = [self.delta, ]
        
        if parameter_ranges is not None:
            params['E0'].append(parameter_ranges[:,0])
            params['E1'].append(parameter_ranges[:,1])
            params['E2'].append(parameter_ranges[:,2])
            params['E3'].append(parameter_ranges[:,3])
            params['h1'].append(parameter_ranges[:,4])
            params['h2'].append(parameter_ranges[:,5])
            params['C1'].append(parameter_ranges[:,6])
            params['C2'].append(parameter_ranges[:,7])
            if self.variant == "kappa":
                params['kappa'].append(parameter_ranges[:,8])
            elif self.variant == "delta":
                params['delta'].append(parameter_ranges[:,8])
            elif self.variant == "both":
                params['kappa'].append(parameter_ranges[:,8])
                params['delta'].append(parameter_ranges[:,9])

        return params
    
    def summary(self, confidence_interval=95, tol=0.01):
        pars = self.get_parameters(confidence_interval=confidence_interval)
        if pars is None:
            return None
        
        ret = []
        key = "kappa"
        if key in pars:
            l = pars[key]
            if len(l)==1:
                if l[0] < -tol:
                    ret.append("%s\t%0.2f\t(<0) antagonistic"%(key, l[0]))
                elif l[0] > tol:
                    ret.append("%s\t%0.2f\t(>0) synergistic"%(key, l[0]))
            else:
                v = l[0]
                lb,ub = l[1]
                if v < -tol and lb < -tol and ub < -tol:
                    ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<0) antagonistic"%(key, v,lb,ub))
                elif v > tol and lb > tol and ub > tol:
                    ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>0) synergistic"%(key, v,lb,ub))

        key = "delta"
        if key in pars:
            l = pars[key]
            if len(l)==1:
                if np.log10(l[0]) < -tol:
                    ret.append("%s\t%0.2f\t(<1) antagonistic"%(key, l[0]))
                elif np.log10(l[0]) > tol:
                    ret.append("%s\t%0.2f\t(>1) synergistic"%(key, l[0]))
            else:
                v = l[0]
                lb,ub = l[1]
                if np.log10(v) < -tol and np.log10(lb) < -tol and np.log10(ub) < -tol:
                    ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(<1) antagonistic"%(key, v,lb,ub))
                elif np.log10(v) > tol and np.log10(lb) > tol and np.log10(ub) > tol:
                    ret.append("%s\t%0.2f\t(%0.2f,%0.2f)\t(>1) synergistic"%(key, v,lb,ub))

        if len(ret)>0:
            return "\n".join(ret)
        else:
            return "No synergy or antagonism detected with %d percent confidence interval"%(int(confidence_interval))

    def __repr__(self):
        if not self._is_parameterized(): return "BRAID()"

        if self.variant == "kappa":
            return "BRAID(E0=%0.2f, E1=%0.2f, E2=%0.2f, E3=%0.2f, h1=%0.2f, h2=%0.2f, C1=%0.2e, C2=%0.2e, kappa=%0.2f)"%(self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.kappa)
        elif self.variant == "delta":
            return "BRAID(E0=%0.2f, E1=%0.2f, E2=%0.2f, E3=%0.2f, h1=%0.2f, h2=%0.2f, C1=%0.2e, C2=%0.2e, delta=%0.2f)"%(self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.delta)
        elif self.variant == "both":
            return "BRAID(E0=%0.2f, E1=%0.2f, E2=%0.2f, E3=%0.2f, h1=%0.2f, h2=%0.2f, C1=%0.2e, C2=%0.2e, kappa=%0.2f, delta=%0.2f)"%(self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.kappa, self.delta)


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

class MarginalLinear:
    """Some drugs' dose response may not follow some known parametric equation. In these cases, MarginalLinear() fits a piecewise linear model mapping log(d) -> E

    Parameters
    ----------
    aggregation_function : function, default=np.mean
        If d contains repeated values (e.g., experimental replicates using the same dose many times), E at these repeated doses will be averaged using aggregation_function(E[d==X])

    strict_inverse : bool, default=False
        Dose response curves are typically strictly monotonic, however due to measurement noise or intrinsic biology, a piecewise linear model of the dose-response curve may not be strictly monotonic (e.g., non-invertible). If strict_inverse==True, all E's that do not correspond to a unique d will return E_inv(E)=NaN. If strict_inverse=False, non-monotonic regions will be cut out and replaced with a monotonic, log-linear curve so that E_inv can be uniquely calculated.
    """

    # NOTE on __init__() and fit(): **kwargs is only present to avoid errors
    # people may run into when reusing code for fitting Hill functions in which
    # they set kwargs. There are no kwargs used for these methods
    def __init__(self, aggregation_function=np.mean, strict_inverse=False, **kwargs):
        self._d = None
        self._E = None
        self._logd = None
        self._aggregation_function = aggregation_function
        self._fit = False
        self._strict_inverse = strict_inverse

        self.E0 = None
        self.Emax = None
        
        
        # These will be filled based on which regions are invertible
        self._logd_for_inverse = None
        self._E_for_inverse = None
        self._uninvertible_domains = []
        self._ready_for_inverse = False

    def fit(self, d, E, **kwargs):
        """Calls __init__(d, E, aggregation_function)

        Parameters
        ----------
        d : array_like
            Array of doses measured
        
        E : array_like
            Array of effects measured at doses d
        """
        self._fit  = True
        self._ready_for_inverse = False

        if (len(d) > len(np.unique(d))):
            self._d = []
            self._E = []

            # Given repeated dose measurements, average E for each
            for val in np.unique(d):
                self._d.append(val)
                self._E.append(self._aggregation_function(E[d==val]))

            self._d = np.asarray(self._d)
            self._E = np.asarray(self._E)

        else:
            self._d = np.array(d,copy=True)
            self._E = np.array(E,copy=True)

        # Replace 0 doses with the minimum float value
        self._d[self._d==0] = np.nextafter(0,1)

        # Sort doses and E
        sorted_indices = np.argsort(self._d)
        self._d = self._d[sorted_indices]
        self._E = self._E[sorted_indices]

        self.E0 = self._E[0]
        self.Emax = self._E[-1]
        
        # Get log-transformed dose (used for interpolation)
        self._logd = np.log(self._d)

    def E(self, d):
        """Evaluate this model at dose d

        Parameters
        ----------
        d : array_like
            Doses to calculate effect at
        
        Returns
        ----------
        effect : array_like
            Evaluate's the model at dose in d
        """
        if not self._fit: return d*np.nan

        d = np.array(d, copy=True)
        d[d==0] = np.nextafter(0,1)
        logd = np.log(d)

        E = np.interp(logd, self._logd, self._E, left=np.nan, right=np.nan)

        return E

    def E_inv(self, E):
        """Find the dose that will achieve effects in E. Only works for dose responses with strictly increasing or strictly decreasing E.

        Parameters
        ----------
        E : array_like
            Effects to get the doses for
        
        Returns
        ----------
        doses : array_like
            Doses which achieve effects E using this model.
        """
        if not self._fit: return E*np.nan
        if not self._ready_for_inverse: self._prepare_inverse()
        if len(self._logd_for_inverse)==0: return E*np.nan

        # These effects are below the minimum or above the maximum, and therefore cannot be used for interpolation
        invalid_mask = (E < min(self._E)) | (E > max(self._E))

        
        if self._strict_inverse:
            # Any E inside a domain we have found to be un-invertible is also invalid. If _strict_inverse is False (default), the entire un-invertible domain will just be filled with a straight line. In this region, the model would have E_inv(E(d))!=d. If strict_inverse is True, this whole region is exlucded (np.nan)
            for domain in self._uninvertible_domains:
                a,b = sorted(domain)
                invalid_mask = invalid_mask | ((E > a) & (E < b))
        
        d = np.exp(np.interp(E, self._E_for_inverse, self._logd_for_inverse))

        if hasattr(E, "__iter__"):
            invalid_mask = np.where(invalid_mask)
            d[invalid_mask] = np.nan
        elif invalid_mask:
            d = np.nan

        return d

    def create_fit(d, E, aggregation_function=np.mean):
        """Courtesy function to build a marginal linear model directly from 
        data. Initializes a model using the provided aggregation function, then 
        fits.
        """
        drug = MarginalLinear(aggregation_function=aggregation_function)
        drug.fit(d, E)
        return drug

    def is_fit(self):
        return self._fit


    def _prepare_inverse(self):
        if not self._fit: return
        self._get_invertible_domains(self._E)
        
        # Which data points fall inside which uninvertable domains?
        index_to_udomain = dict()

        # Which data points are outside of uninvertable domains?
        valid_indices = []

        # These are the points we will use for interpolation. We will later extend them by the boundaries of the uninvertible domains
        valid_E = []
        valid_d = []

        # Loop over every point and check whether it is in a u_domain
        for i,E in enumerate(self._E):
            found = False
            for domain in self._uninvertible_domains:
                if E >= min(domain) and E <= max(domain):
                    index_to_udomain[i] = domain
                    found = True
                    break
            if not found:
                valid_indices.append(i)
                valid_E.append(E)
                valid_d.append(self._logd[i])

        # Imagine 6 data points [0, 1, 2, 3, 4, 5]
        # If valid_indices = [0,1,5] then bad_indices = [2, 3, 4]
        # 2 is adjacent to a valid index, so we interpolate between them
        # Likewise for 4
        # General strategy - loop over bad_indices, loop over +/-1 neighbors, if in valid_indices, find boundary point

        for invalid_i in index_to_udomain.keys():
            for neighbor in [-1,1]:
                neighbor_i = invalid_i + neighbor
                if neighbor_i in valid_indices:
                    E = self._E[neighbor_i]
                    ld = self._logd[neighbor_i]

                    E2 = self._E[invalid_i]
                    ld2 = self._logd[invalid_i]
                    domain = index_to_udomain[invalid_i]


                    # Is the valid neigbor above or below the domain?
                    if E < min(domain): # Below
                        E_target = min(domain)
                        Elower = E
                        Eupper = E2
                        dlower = ld
                        dupper = ld2
                    else: # above
                        E_target = max(domain)
                        Elower = E2
                        Eupper = E
                        dlower = ld2
                        dupper = ld
                    logd_target = np.interp(E_target, [Elower, Eupper], [dlower, dupper])

                    # These are new points that exist along the log-linear interpolation, where (d,E) exist at the boundary of un-invertible domains
                    valid_d.append(logd_target)
                    valid_E.append(E_target)

        # What about uninvertible domains that don't touch, but have no good
        # point between them? E.g., E3 is in (0.5,0.6), and E4 is in (0.7,0.8)
        #  - then we need to create two new points - at the inner boundaries
        #    of both domains
        for _i in index_to_udomain.keys():
            dom_i = index_to_udomain[_i]
            
            _j = _i + 1
            if _j in index_to_udomain:
                dom_j = index_to_udomain[_j]
                if dom_j == dom_i: continue # They are in the same domain

                E_i = self._E[_i] # This is inside dom_i
                ld_i = self._logd[_i]
                E_j = self._E[_j] # This is inside dom_j
                ld_j = self._logd[_j]

                if E_i < E_j:
                    Elower = E_i
                    Eupper = E_j
                    dlower = ld_i
                    dupper = ld_j
                    E_boundary_i = max(dom_i)
                    E_boundary_j = min(dom_j)
                else:
                    Elower = E_j
                    Eupper = E_i
                    dlower = ld_j
                    dupper = ld_i
                    E_boundary_j = max(dom_j)
                    E_boundary_i = min(dom_i)
                
                ld_boundary_i = np.interp(E_boundary_i, [Elower, Eupper], [dlower, dupper])
                ld_boundary_j = np.interp(E_boundary_j, [Elower, Eupper], [dlower, dupper])

                valid_d.append(ld_boundary_i)
                valid_E.append(E_boundary_i)
                valid_d.append(ld_boundary_j)
                valid_E.append(E_boundary_j)

        
        sorted_indices = np.argsort(valid_E)
        
        self._E_for_inverse = np.asarray(valid_E)[sorted_indices]
        self._logd_for_inverse = np.asarray(valid_d)[sorted_indices]
        self._ready_for_inverse = True

    def _get_invertible_domains(self, E):
        if not self._fit: return
        self._uninvertible_domains = []
        for i in range(len(E)-2):
            for j in range(i+1, len(E)-1):
                r1 = [E[i], E[i+1]]
                r2 = [E[j], E[j+1]]
                overlap = self._interval_intersection(r2,r1)
                if overlap is not None:
                    # Does this overlap with any regions we have seen before?
                    previous_overlaps = [] # We will merge these all together
                    for _i, prev_olap in enumerate(self._uninvertible_domains):
                        overlap_2 = self._interval_intersection(overlap, prev_olap, border_is_overlap=True)
                        if overlap_2 is not None:
                            previous_overlaps.append(prev_olap)

                    if len(previous_overlaps)==0: # No previous overlap
                        self._uninvertible_domains.append(overlap)
                    else: # Merge this set with all previous overlaps
                        for po in previous_overlaps: # remove previous ones
                            self._uninvertible_domains.remove(po)
                        self._uninvertible_domains.append(self._interval_union_multiple(previous_overlaps + [overlap,]))

    def _interval_intersection(self, r1, r2, border_is_overlap=False):
        r1 = sorted(r1) # low -> high
        r2 = sorted(r2) # low -> high
        
        # If the low value of one interval is above the high value of the other, there can be no overlap
        if not border_is_overlap:
            if (r2[0] >= r1[1] or r1[0] >= r2[1]): return None
        else:
            if (r2[0] > r1[1] or r1[0] > r2[1]): return None

        # The low value is the max of the two low values, and the high value is the min of the two high values
        return (max(r1[0],r2[0]), min(r1[1],r2[1]))

    def _interval_union(self, r1, r2):
        r1 = sorted(r1)
        r2 = sorted(r2)
        return (min(r1[0], r2[0]), max(r1[1], r2[1]))

    def _interval_union_multiple(self, sets):
        if len(sets)==0: return None
        if len(sets)==1: return sets[0]
        r1 = sets[0]
        for r2 in sets[1:]:
            r1 = self._interval_union(r1, r2)
        return r1


#    Copyright (C) 2020 David J. Wooten
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import inspect
import warnings


def sham(d, drug):
    """Simulates a sham combination experiment. In a sham experiment, the two drugs combined are (secretly) the same drug. For example, a sham combination may add 10uM drugA + 20uM drugB. But because drugA and drugB are the same (drugX), the combination is really just equivalent to 30uM of the drug.    
    """
    if not 0 in d:
        d = np.append(0,d)
    d1, d2 = np.meshgrid(d,d)
    d1 = d1.flatten()
    d2 = d2.flatten()
    E = drug.E(d1+d2)
    return d1, d2, E

def sham_higher(d, drug, n_drugs):
    """Simulates a sham combination experiment for 3+ drugs. In a sham experiment, the two drugs combined are (secretly) the same drug. For example, a sham combination may add 10uM drugA + 20uM drugB. But because drugA and drugB are the same (drugX), the combination is really just equivalent to 30uM of the drug.

    Parameters
    ----------
    d : array_like
        Dose escalation to use for each "sham" drug

    drug
        A parameterized drug model from synergy.single, such as Hill, Hill_2P, Hill_CI, or MarginalLinear.

    n_drugs : int
        The number of drugs to include in the sham combination

    Returns
    ----------
    doses : M x n_drugs numpy.ndarray
        All dose pairs for each combination. In total, there are M samples, taken from n_drugs drugs.

    E : numpy.array
        Sham effects calculated as drug.E(doses.sum(axis=1))
    """
    if not 0 in d:
        d = np.append(0,d)
    doses = [d, ]*n_drugs
    doses = list(np.meshgrid(*doses))
    for i in range(n_drugs):
        doses[i] = doses[i].flatten()
    doses = np.asarray(doses).T
    E = drug.E(doses.sum(axis=1))
    return doses, E

def remove_zeros(d, min_buffer=0.2):
    """Replace zeros with some semi-intelligently chosen small value

    When plotting on a log scale, 0 doses can cause problems. This replaces all 0's using the dilution factor between the smallest non-zero, and second-smallest non-zero doses. If that dilution factor is too close to 1, it will replace 0's doses with a dose that is min_buffer*(max(d)-min(d[d>0])) less than min(d[d>0]) on a log scale.

    Parameters
    ----------
    d : array_like
        Doses to remove zeros from. Original array will not be changed.

    min_buffer : float , default=0.2
        For very large dose arrays with very small step sizes (useful for getting smooth plots), replacing 0's may lead to a value too close to the smallest non-zero dose. min_buffer is the minimum buffer (in log scale, relative to the full dose range) that 0's will be replaced with.
    """

    d=np.array(d,copy=True)
    dmin = np.min(d[d>0]) # smallest nonzero dose
    dmin2 = np.min(d[d>dmin])
    dilution = dmin/dmin2

    dmax = np.max(d)
    logdmin = np.log(dmin)
    logdmin2 = np.log(dmin2)
    logdmax = np.log(dmax)

    if (logdmin2-logdmin) / (logdmax-logdmin) < min_buffer:
        logdmin2_effective = logdmin + min_buffer*(logdmax-logdmin)
        dilution = dmin/np.exp(logdmin2_effective)

    d[d==0]=dmin * dilution
    return d

def residual_ss(d1, d2, E, model):
    E_model = model(d1, d2)
    return np.sum((E-E_model)**2)

def residual_ss_1d(d, E, model):
    E_model = model(d)
    return np.sum((E-E_model)**2)

def AIC(sum_of_squares_residuals, n_parameters, n_samples):
    """
    SOURCE: AIC under the Framework of Least Squares Estimation, HT Banks, Michele L Joyner, 2017
    Equations (6) and (16)
    https://projects.ncsu.edu/crsc/reports/ftp/pdf/crsc-tr17-09.pdf
    """
    aic = n_samples * np.log(sum_of_squares_residuals / n_samples) + 2*(n_parameters + 1)
    if n_samples / n_parameters > 40:
        return aic
    else:
        return aic + 2*n_parameters*(n_parameters+1) / (n_samples - n_parameters - 1)

def BIC(sum_of_squares_residuals, n_parameters, n_samples):
    return n_samples * np.log(sum_of_squares_residuals / n_samples) + (n_parameters+1)*np.log(n_samples)

def r_squared(E, sum_of_squares_residuals):
    ss_tot = np.sum((E-np.mean(E))**2)
    return 1-sum_of_squares_residuals/ss_tot

def sanitize_initial_guess(p0, bounds):
    """
    Makes sure p0 is within the bounds
    """
    index = 0
    for x, lower, upper in zip(p0, *bounds):
        if x is None:
            if True in np.isinf((lower,upper)): np.min((np.max((0,lower)), upper))
            else: p0[index]=np.mean((lower,upper))

        elif x < lower: p0[index]=lower
        elif x > upper: p0[index]=upper
        index += 1

def sanitize_single_drug_model(model, default_class, expected_superclass=None, **kwargs):
    """
    Makes sure the given single drug model is a class or object of a class that is permissible for the given synergy model.

    Parameters
    ----------
    model : object or class
        A single drug model

    default_class : class
        The type of model to return if the given model is of the wrong type

    expected_superclass : class , default=None
        The class the model is expected to be an instance of

    Returns
    -------
    model : object
        An object that is an instance of expected_superclass
    """
    # The model is a class
    if inspect.isclass(model):

        # If there is no expected_superclass, assume the given class is fine
        if expected_superclass is None:
            return model(**kwargs)

        else:
            # The model is a subclass of the expected subclass
            if issubclass(model, expected_superclass):
                # We are good!
                return model(**kwargs)

            # The given class violates the expected class: return the default
            else:
                if model is not None: warnings.warn("Expected single drug model to be subclass of %s, instead got %s"%(expected_superclass, model))
                return default_class(**kwargs)
        
        return model(**kwargs)

    # The model is an object
    else:
        # There is no expected_superclass, so assume the object is fine
        if expected_superclass is None:
            if model is None:
                return default_class(**kwargs)
            return model
        
        # The model is an instance of the expected_superclass, so good!
        elif isinstance(model,expected_superclass):
            return model

        # The model is an instance of the wrong type of class, so return the default
        else:
            if model is not None: warnings.warn("Expected single drug model to be subclass of %s, instead got %s"%(expected_superclass, type(model)))

            return default_class(**kwargs)
    return default_class(**kwargs)
